/* sdramtest.c - a test program that loads an image from an SD card and runs a video demo */

/* the SDRAM driver is by cgracey */
/* The VGA and Wired drivers are by Baggers */
/* the C code is by David Betz */

/* before running the Makefile you must build the three drivers using PNut.exe */

#include <stdio.h>
#include <stdint.h>
#include <dirent.h>
#include <propeller.h>
#include <sys/sd.h>

#define LEN_16      0x00000001
#define LEN_32      0x00000002
#define LEN_64      0x00000003
#define LEN_128     0x00000004
#define LEN_256     0x00000005
#define LEN_512     0x00000006
#define LEN_1024    0x00000007

#define READ        0x00000000
#define WRITE       0x00000008

#define MEMORY_SIZE             (16 * 1024 * 1024)
#define BLOCK_SIZE              64
#define BLOCK_SIZE_IN_LONGS     (BLOCK_SIZE / sizeof(uint32_t))
#define ALIGN                   16
#define FRAME_SIZE              (1024 * 600)

#define FILE_NAME               "WW1024~1.BIN" // "ww1024x600.bin"

typedef struct {
    volatile uint32_t **p_vbptr;
    volatile uint32_t *p_vidramptr;
    volatile uint32_t *sdram_cmd_entry;
    void *hubram_addr;
} VGA_PARAMS;

typedef struct {
    volatile uint32_t **p_vbptr;
    volatile uint32_t *p_vidramptr;
    volatile uint32_t *sdram_cmd_entry;
    void *workram;
    uint32_t sdram_frame0;
    uint32_t sdram_frame1;
} WIRED_PARAMS;

extern _Driver _SimpleSerialDriver;
extern _Driver _FileDriver;

_Driver *_driverlist[] = {
  &_SimpleSerialDriver,
  &_FileDriver,
  NULL
};

#define DO   	44
#define DI      45
#define CLK     46
#define CS      47

void LoadSDDriver(uint32_t configwords[]);

int main (int argc,  char* argv[])
{
    extern char binary_SDRAM_Driver_obj_start[];
    extern char binary_VGA_Driver_obj_start[];
    extern char binary_Wired_Driver_obj_start[];
    uint32_t addr, data, data_start;
    int count, errors, i;
    FILE *fp;
    
    /* variables shared between the VGA driver and the Wired driver */
    volatile uint32_t *vbptr;
    volatile uint32_t vidramptr = 0x00000000;
    volatile uint32_t sdram_cmds[5] = { 0, 0, 0, 0, 8 };

    /* create a 16 byte aligned buffer for loading the image */
    uint8_t buffer[BLOCK_SIZE + ALIGN];
    uint32_t *bufferp = (uint32_t *)(((uint32_t)buffer + ALIGN) & ~(ALIGN - 1));
    
    /* create a 16 byte aligned workspace for the VGA driver */
    uint8_t vgawork[4096 + ALIGN];
    uint8_t *vgaworkp = (uint8_t *)(((uint32_t)vgawork + ALIGN) & ~(ALIGN - 1));
    
    /* create a 16 byte aligned workspace for the Wired driver */
    uint8_t workram[4*1024 + ALIGN];
    uint8_t *workramp = (uint8_t *)(((uint32_t)workram + ALIGN) & ~(ALIGN - 1));

    /* parameters to initialize the VGA driver */
    VGA_PARAMS vga_params = { &vbptr, &vidramptr, &sdram_cmds[2], vgaworkp };

    /* parameters to initialize the Wired driver */
    WIRED_PARAMS wired_params = { &vbptr, &vidramptr, &sdram_cmds[0], workramp, 0, FRAME_SIZE };

    /* setup the pins used by the SD card driver */
    _SD_Params sd_config;
    sd_config.AttachmentType = _SDA_ConfigWords;
    sd_config.pins.ConfigWords.CONFIG1 = (DI<<24) | (DO<<16) | (CLK<<8) | 1;
    sd_config.pins.ConfigWords.CONFIG2 = CS<<24;
    
    /* start the SDRAM driver so we can load the image from the SD card */
    printf("Starting SDRAM driver\n");
    coginit(2, binary_SDRAM_Driver_obj_start, (void *)sdram_cmds);
    
    /* mount the SD card */
    /* This is done explicitly because p2load doesn't know how to patch configuration variables yet */
    if (dfs_mount(&sd_config) != 0) {
        printf("SD mount failed\n");
        return 1;
    }

    /* open the image file on the SD card */
    printf("Opening '%s'\n", FILE_NAME);
    if ((fp = fopen(FILE_NAME, "rb")) == NULL) {
        printf("Open of '%s' failed\n", FILE_NAME);
        return 1;
    }
    
    /* load the image into SDRAM and close the file */
    printf("Loading image..."); fflush(stdout);
    addr = 0x00000000;
    while (fread(bufferp, 1, BLOCK_SIZE, fp) == BLOCK_SIZE) {
        sdram_cmds[1] = addr;
        sdram_cmds[0] = (uint32_t)bufferp | LEN_64 | WRITE;
        while (sdram_cmds[0])
            ;
        addr += BLOCK_SIZE;
    }
    printf("done\n");
    fclose(fp);
    
    /* finally, start the VGA driver in COG 0 since that is required on the DE2-115 emulation */
    printf("Starting VGA driver\n");
    coginit(0, binary_VGA_Driver_obj_start, &vga_params);
    
    /* restart the C COG with the Wired driver */
    printf("Starting Wired driver\n");
    coginit(cogid(), binary_Wired_Driver_obj_start, &wired_params);
    
    /* at this point the C code will no longer be running and only Baggers' and Chip's drivers will be running */
    /* so this point is never reached */
    printf("Should never get here!\n");
    
    return 0;
}
